import os
import sys

import sounddevice as sd
import soundfile as sf


class AudioPlayer:
    """
    A simple class for playing WAV audio files at specified timings
    """

    def __init__(self):
        """Initialize the audio player"""
        self.audio_cache = {}  # Cache for loaded audio data

    def load_audio(self, sound_name):
        """
        Load a WAV file and save it to cache

        Args:
            filepath (str): Path to the WAV file

        Returns:
            tuple: (audio data, sample rate)
        """
        if getattr(sys, "frozen", False):
            current_dir = os.path.dirname(sys.executable)
        else:
            current_dir = os.path.dirname(__file__)
        sound_path = os.path.join(current_dir, "sound", sound_name)

        if sound_path not in self.audio_cache:
            data, samplerate = sf.read(sound_path)
            self.audio_cache[sound_path] = (data, samplerate)
        return self.audio_cache[sound_path]

    def play_sound(self, filepath, play_time=None):
        """
        Play audio at a specified time

        Args:
            filepath (str): Path to the WAV file to play
            play_time (datetime.datetime, optional): Time to play. If None, play immediately
        """
        # Pre-load the audio
        data, samplerate = self.load_audio(filepath)

        if play_time is None:
            # Play immediately
            sd.play(data, samplerate)
            return



import threading
import time

from . import ble_control


class Trigger:
    """録画トリガーの公開インターフェースは変更しない。
    set_device_names と set_keyword と set_delay_sec
    ble_connect と trigger_in と trigger_out
    get_triggered と set_status と get_status
    """

    def __init__(self, offset_sec: int = 5) -> None:
        self.triggered_in = False
        self.offset_sec = offset_sec
        self.keyword = "(recording)"
        self.ble_thread = ble_control.BleThread()
        self.target_device_names = []
        self.connection_status = "Idle"
        self.delay_sec = 1

    def ble_connect(self) -> None:
        self.ble_thread.start()
        command, ok_count, msg = self.ble_thread.execute_command(
            "connect", None, timeout=30
        )
        if ok_count == len(self.target_device_names):
            self.connection_status = "Connected"
        else:
            self.connection_status = (
                f"Failed ({ok_count}/{len(self.target_device_names)})"
            )
            print(f"BLE connect failed: {msg}")

    def set_device_names(self, names):
        self.target_device_names = names
        self.ble_thread.set_target_device_names(names)

    def set_keyword(self, keyword):
        self.keyword = keyword

    def set_delay_sec(self, delay_sec):
        self.delay_sec = delay_sec

    def trigger_in(self, title):
        if self.keyword in title and self.triggered_in is False:
            self.triggered_in = True
            command, success, msg = self.ble_thread.execute_command(
                "record_start", None, timeout=3
            )
            if success:
                self.connection_status = "Recording"
                return True
            self.connection_status = "Failed to start"
        return False

    def trigger_out(self, title):
        if self.keyword not in title and self.triggered_in is True:
            print("trigger out")
            self.triggered_in = False
            threading.Thread(target=self._delayed_stop, daemon=True).start()
            return True
        return False

    def _delayed_stop(self):
        time.sleep(self.delay_sec)
        command, success, msg = self.ble_thread.execute_command(
            "record_stop", None, timeout=5
        )
        if success:
            self.connection_status = "Connected"
        else:
            self.connection_status = "Failed to stop"

    def get_triggered(self):
        return self.triggered_in

    def set_status(self, status):
        self.connection_status = status

    def get_status(self):
        return self.connection_status

    def update_status(self):
        cmd, alive_cnt, msg = self.ble_thread.execute_command("status", None, timeout=3)
        if cmd != "status":
            return self.connection_status
        try:
            ratio, state = msg.split(maxsplit=1)
            alive_str, total_str = ratio.split("/")
            total = int(total_str)
        except Exception:
            return self.connection_status

        if total == 0:
            # 未接続またはデバイス未検出
            return self.connection_status

        if alive_cnt == 0:
            self.connection_status = "Disconnected"
        elif alive_cnt < total:
            self.connection_status = f"KeepAlive Failed ({alive_cnt}/{total})"
        else:
            if state.startswith("recording"):
                self.connection_status = "Recording..."
            else:
                self.connection_status = "Connected"
        return self.connection_status



def timedelta_to_str(td, hhmmss=True):
    if hhmmss:
        return timedelta_to_str_hh_mm_ss(td)
    else:
        return timedelta_to_str_mm_ss(td)


def timedelta_to_str_hh_mm_ss(td):
    hours = td.seconds // 3600
    remain = td.seconds - (hours * 3600)
    minutes = remain // 60
    seconds = remain - (minutes * 60)
    return f"{int(hours)}:{int(minutes):02}:{int(seconds):02}"


def timedelta_to_str_mm_ss(td):
    minutes = td.seconds // 60
    seconds = td.seconds - (minutes * 60)
    return f"{int(minutes)}:{int(seconds):02}"


def seconds_to_time_str(src):
    if isinstance(src, str):
        if src == "":
            return ""
        colons = src.count(":")
        if colons == 2:
            return src
        elif colons == 1:
            minutes, seconds = src.split(":")
            hours = int(minutes) // 60
            minutes = int(minutes) % 60
            return f"{hours}:{minutes:02}:{int(seconds):02}"
        elif colons == 0:
            sec = int(src)
    else:
        sec = src
    hours = sec // 3600
    remain = sec - (hours * 3600)
    minutes = remain // 60
    seconds = remain - (minutes * 60)
    return f"{int(hours)}:{int(minutes):02}:{int(seconds):02}"


def time_str_to_seconds(time_str):
    if time_str == "":
        time_str = "0"
    colons = time_str.count(":")
    if colons == 2:
        hours, minutes, seconds = time_str.split(":")
        return int(hours) * 3600 + int(minutes) * 60 + int(seconds)
    elif colons == 0:
        return int(time_str)
    minutes, seconds = time_str.split(":")
    return int(minutes) * 60 + int(seconds)



import importlib.util
import os
import shutil
import subprocess
import sys
import tkinter as tk
from datetime import datetime, timedelta
from tkinter import filedialog, ttk

import ttkthemes

from . import file_loader, gui_canvas, gui_tree, icon_data, svg_writer, time_format
from .gui_parts import Combobox

IS_DARWIN = sys.platform.startswith("darwin")


class App(ttk.Frame):
    def __init__(self, master):
        super().__init__(master)
        master.title("Timeline-kun")

        head_frame = ttk.Frame(master)
        head_frame.pack(padx=10, pady=(15, 5), fill=tk.X)
        create_file_btn = ttk.Button(
            head_frame, text="Create CSV", width=13, command=self.create_file
        )
        create_file_btn.pack(padx=5, side=tk.LEFT)
        load_file_btn = ttk.Button(
            head_frame, text="Load CSV", width=13, command=self.select_file
        )
        load_file_btn.pack(padx=5, side=tk.LEFT)
        self.file_path_label = ttk.Label(head_frame, text="No file selected")
        self.file_path_label.pack(padx=5, side=tk.LEFT)

        self.timer_btn = ttk.Button(
            head_frame, text="Send to timer", command=self.open_timer
        )
        self.timer_btn.pack(padx=5, side=tk.RIGHT)
        self.timer_btn["state"] = "disabled"

        values = ["orange", "cyan", "lightgreen"]
        self.timer_color_combobox = Combobox(head_frame, "Color:", values, width=12)
        self.timer_color_combobox.pack_horizontal(padx=5, side=tk.RIGHT)

        send_timer_frame = ttk.Frame(master)
        send_timer_frame.pack(padx=10, pady=(5, 10), fill=tk.X)

        excel_btn = ttk.Button(
            send_timer_frame, text="Send to Excel", width=13, command=self.open_excel
        )
        excel_btn.pack(padx=5, side=tk.LEFT)
        reload_btn = ttk.Button(send_timer_frame, text="Reload", command=self.load_file)
        reload_btn.pack(padx=5, side=tk.LEFT)

        values = ["mm:ss", "h:mm:ss"]
        self.time_format_combobox = Combobox(
            send_timer_frame, "Format", values, width=10
        )
        self.time_format_combobox.pack_horizontal(padx=(70, 5), side=tk.LEFT)
        self.time_format_combobox.set_selected_bind(lambda e: self.draw_stages())

        values = ["tiny", "small", "normal"]
        self.font_size_combobox = Combobox(
            send_timer_frame, "Font size", values, width=10, current=1
        )
        self.font_size_combobox.pack_horizontal(padx=5, side=tk.LEFT, anchor=tk.E)
        self.font_size_combobox.set_selected_bind(lambda e: self.draw_stages())

        values = ["vertical", "horizontal"]
        self.direction_combobox = Combobox(
            send_timer_frame, "Direction", values, width=10
        )
        self.direction_combobox.pack_horizontal(padx=5, side=tk.LEFT, anchor=tk.E)
        self.direction_combobox.set_selected_bind(lambda e: self.draw_stages())

        export_svg_btn = ttk.Button(
            send_timer_frame, text="Export SVG", command=self.export_svg
        )
        export_svg_btn.pack(padx=5, side=tk.LEFT)

        # error message frame
        msg_frame = ttk.Frame(send_timer_frame)
        msg_frame.pack(padx=10, fill=tk.X, side=tk.LEFT)
        self.msg_label = ttk.Label(msg_frame, text="")
        self.msg_label.pack()

        # body frame
        body_frame = ttk.Frame(master)
        body_frame.pack(fill=tk.BOTH, expand=True)
        # tree
        tree_frame = ttk.Frame(body_frame, width=400)
        tree_frame.pack(side=tk.LEFT, fill=tk.Y)
        cols = [
            {"name": "title", "width": 200},
            {"name": "member", "width": 100},
            {"name": "start", "width": 50},
            {"name": "end", "width": 50},
            {"name": "duration", "width": 50},
            {"name": "fixed", "width": 60},
            {"name": "instruction", "width": 200},
        ]
        self.tree = gui_tree.Tree(tree_frame, columns=cols)
        self.tree.pack(fill=tk.BOTH, expand=True)
        self.tree.tree.bind("<<TreeviewSelect>>", lambda e: self.select_row())
        self.tree.add_menu("Set start point", self.draw_start_line)
        self.tree.add_menu("Insert", self.insert_row)
        self.tree.add_menu("Edit", self.edit_row)
        self.tree.add_menu("Remove", self.remove_row)
        # canvas
        canvas_frame = ttk.Frame(body_frame)
        canvas_frame.pack(fill=tk.BOTH, expand=True)
        self.canvas = gui_canvas.Canvas(canvas_frame, bg="white")
        self.canvas.pack(fill=tk.BOTH, expand=True)

        # ctrl+z shortcut control
        master.bind("<Control-z>", self.undo)

        self.csv_path = None
        self.start_index = 0
        self.prev_end_time_sec = None
        self.next_start_time_sec = None

    def edit_row(self):
        if self._check_csv_file_locked(self.csv_path) is True:
            return

        is_ok = self.tree.edit(self.prev_end_time_sec)
        if is_ok:
            self.tree.tree_to_csv_file(self.csv_path)
            self.load_file()

    def insert_row(self):
        if self._check_csv_file_locked(self.csv_path) is True:
            return
        is_ok = self.tree.insert()
        if is_ok:
            self.tree.tree_to_csv_file(self.csv_path)
            self.load_file()

    def remove_row(self):
        if self._check_csv_file_locked(self.csv_path) is True:
            return
        # confirm remove
        current_row = self.tree.get_selected_index()
        if current_row is None:
            return
        current_title = self.stage_list[current_row]["title"]
        is_ok = tk.messagebox.askyesno(
            "Confirm",
            f"Do you want to remove the row ({current_title})?.",
        )
        if is_ok is False:
            return
        is_ok = self.tree.remove()

        if is_ok:
            self.tree.tree_to_csv_file(self.csv_path)
            self.load_file()

    def _check_csv_file_locked(self, file_path):
        if self.tree.check_csv_file_locked(self.csv_path) is True:
            tk.messagebox.showinfo(
                "Info",
                f"The file is currently locked by another application, such as Excel.\n{self.csv_path}",
            )
            return True
        return False

    def select_row(self):
        idx = self.tree.get_selected_index()
        if idx is None:
            return
        prev_fixed_code, next_fixed_code = self._get_prev_next_fixed_code(idx)

        # preview event
        if prev_fixed_code is None:
            self.prev_end_time_sec = self._get_prev_start_time(idx)
        elif prev_fixed_code == "start":
            prev_end_sec = self.stage_list[idx - 1]["end_sec"]
            prev_duration_sec = self.stage_list[idx - 1]["duration_sec"]
            # end_sec and duration_sec are 0 -> prevprev
            if prev_end_sec == "0" and prev_duration_sec == "0":
                self.prev_end_time_sec = self._get_prev_start_time(idx) + 1
            else:
                self.prev_end_time_sec = self._get_prev_end_time(idx)
        elif prev_fixed_code == "duration":
            self.prev_end_time_sec = self._get_prev_end_time(idx)

        # next event
        if next_fixed_code == "start":
            self.next_start_time_sec = self._get_next_start_time(idx)
        elif next_fixed_code == "duration":
            self.next_start_time_sec = self._get_next_end_time(idx) - 1
        elif next_fixed_code is None:
            self.next_start_time_sec = self._get_next_end_time(idx)

        self.highlight_selected_row()

    def draw_start_line(self):
        self.start_index = self.tree.get_selected_index()
        if self.start_index is None:
            self.start_index = 0

        stage = self.stage_list[self.start_index]
        total_duration = self.stage_list[-1]["end_dt"].total_seconds()
        self.canvas.draw_start_line(stage["start_dt"], total_duration)

        for s in self.stage_list:
            caption = self._minus_timedelta(s["start_dt"], stage["start_dt"])
            self.canvas.create_time(s["start_dt"], text=caption)
        total_end_dt = self.stage_list[-1]["end_dt"]
        time_caption = self._minus_timedelta(total_end_dt, stage["start_dt"])
        self.canvas.create_time(total_end_dt, text=time_caption)

    def highlight_selected_row(self):
        idx = self.tree.get_selected_index()
        if idx is None:
            return
        start = self.stage_list[idx]["start_dt"]
        duration = self.stage_list[idx]["duration"]
        self.canvas.delete("highlight")
        self.canvas.create_rect(start, duration, "#dd7777", tag="highlight")

    def _get_prev_next_fixed_code(self, index):
        if index == 0:
            prev_fixed_code = None
        else:
            prev_fixed_code = self.stage_list[index - 1]["fixed"]
        if index == len(self.stage_list) - 1:
            next_fixed_code = None
        else:
            next_fixed_code = self.stage_list[index + 1]["fixed"]
        return prev_fixed_code, next_fixed_code

    def _get_prev_end_time(self, index):
        if index <= 0:
            return 0
        else:
            return self.stage_list[index - 1]["end_dt"].total_seconds()

    def _get_prev_start_time(self, index):
        if index <= 0:
            return 0
        else:
            return self.stage_list[index - 1]["start_dt"].total_seconds()

    def _get_next_start_time(self, index):
        if index >= len(self.stage_list) - 1:
            return 1000000
        else:
            return self.stage_list[index + 1]["start_dt"].total_seconds()

    def _get_next_end_time(self, index):
        if index >= len(self.stage_list) - 1:
            return 1000000
        else:
            return self.stage_list[index + 1]["end_dt"].total_seconds()

    def _minus_timedelta(self, td_1, td_2):
        include_hour = self.time_format_combobox.get() == "h:mm:ss"
        if td_1 < td_2:
            total_seconds = td_2 - td_1
            return f"-{time_format.timedelta_to_str(total_seconds, include_hour)}"
        else:
            total_seconds = td_1 - td_2
            return time_format.timedelta_to_str(total_seconds, include_hour)

    def draw_stages(self):
        self.canvas.delete("all")
        self.canvas.set_font(self.font_size_combobox.get())
        self.canvas.set_direction(self.direction_combobox.get())
        total_duration = self.stage_list[-1]["end_dt"].total_seconds()
        self.canvas.set_scale(total_duration)

        past_rect_height = 10000
        include_hour = self.time_format_combobox.get() == "h:mm:ss"

        past_start_dt = timedelta(seconds=0)
        for stage in self.stage_list:
            if stage["start_dt"].total_seconds() == 0:
                start_dt = past_start_dt
            else:
                start_dt = stage["start_dt"]
            rect_height = self.canvas.create_rect(
                start_dt, stage["duration"], stage["color"]
            )
            label_title = f"{stage['title']}"
            label_time = (
                f"{time_format.timedelta_to_str(stage['duration'], include_hour)}"
            )
            self.canvas.create_label(
                start_dt, stage["duration"], label_title, label_time, stage["has_error"]
            )
            if past_rect_height > 10:
                time_caption = time_format.timedelta_to_str(start_dt, include_hour)
                self.canvas.create_time(stage["start_dt"], text=time_caption)
            past_start_dt = stage["start_dt"]
            past_rect_height = rect_height
        time_caption = time_format.timedelta_to_str(
            self.stage_list[-1]["end_dt"], include_hour
        )
        self.canvas.create_time(self.stage_list[-1]["end_dt"], text=time_caption)

    def create_file(self):
        desktop_path = os.path.join(os.path.expanduser("~"), "Desktop")
        dir_path = filedialog.askdirectory(initialdir=desktop_path)
        if not dir_path:
            return
        today = datetime.now().strftime("%Y-%m-%d")
        file_path = os.path.join(dir_path, f"{today}_new.csv")
        if os.path.exists(file_path):
            tk.messagebox.showinfo("Info", "File already exists.")
            return
        with open(file_path, "w") as f:
            f.write("title,member,start,duration,fixed,instruction\n")
            f.write("TASK A,MEMBER1,,0:01:00,duration,The first task.\n")
            f.write("TASK B,MEMBER1,0:01:10,0:01:30,start,The second task.\n")
            f.write("TASK C,MEMBER1,,0:01:00,duration,The third task.\n")
            f.write("TASK D,MEMBER1,,0:01:00,duration,The 4th task.\n")
            f.write("TASK E,MEMBER1,0:05:00,0:01:00,start,The final task.\n")

        self.csv_path = file_path
        self.load_file()

    def select_file(self):
        self.msg_label.config(text="")
        csv_path = filedialog.askopenfilename(filetypes=[("CSV files", "*.csv")])
        if not csv_path:
            return
        self.csv_path = csv_path
        self.load_file()

    def load_file(self):
        self.start_index = 0
        self.file_path_label.config(text=self.csv_path)

        fl = file_loader.FileLoader()
        try:
            warn_msg, timetable = fl.load_file_for_preview(self.csv_path)
        except Exception as e:
            self.timer_btn["state"] = "disabled"
            self.msg_label.config(text=f"[ERROR]{e}")
            return

        self.stage_list = []
        for row in timetable.get_timetable():
            self.stage_list.append(
                {
                    "title": row["title"],
                    "member": row["member"],
                    "start_dt": row["start"],
                    "end_dt": row["end"],
                    "duration": row["end"] - row["start"],
                    "start_sec": row["start_sec"],
                    "end_sec": row["end_sec"],
                    "duration_sec": row["duration_sec"],
                    "fixed": row["fixed"],
                    "instruction": row["instruction"],
                    "has_error": row["has_error"],
                }
            )
        self.asign_rect_color()
        self.tree.set_stages(self.stage_list)
        self.draw_stages()

        if warn_msg != "":
            self.msg_label.config(text=f"[ERROR]{warn_msg}")
            self.timer_btn["state"] = "disabled"
        else:
            self.msg_label.config(text="Successfully loaded.")
            self.timer_btn["state"] = "normal"

        self.csv_encoding = fl.get_encoding()
        if self.csv_encoding is not None:
            self.tree.set_write_encoding(self.csv_encoding)

    def asign_rect_color(self):
        title_list = list(set([s["title"] for s in self.stage_list]))
        title_list.sort()
        for i, title in enumerate(title_list):
            for stage in self.stage_list:
                if stage["title"] == title:
                    if i >= len(gui_canvas.rect_colors):
                        stage["color"] = "#aaaaaa"
                    else:
                        stage["color"] = gui_canvas.rect_colors[i]

    def open_timer(self):
        # has_error check
        has_error_list = [s["has_error"] for s in self.stage_list]
        if True in has_error_list:
            tk.messagebox.showinfo("Error", "Timer can't start because of error.")
            return

        if self.time_format_combobox.get() == "h:mm:ss":
            hmmss = "hmmss"
        else:
            hmmss = "mmss"

        # frozen exe
        if getattr(sys, "frozen", False):
            current_dir = os.path.dirname(sys.executable)
            tar_path = os.path.join(current_dir, "TimelinekunTimer.exe")
            color = self.timer_color_combobox.get()

            subprocess.Popen(
                [
                    tar_path,
                    "--file_path",
                    self.csv_path,
                    "--fg_color",
                    color,
                    "--start_index",
                    str(self.start_index),
                    "--hmmss",
                    hmmss,
                ]
            )
        elif (
            importlib.util.find_spec(
                f"{__package__}.app_timer" if __package__ else "app_timer"
            )
            is not None
        ):
            color = self.timer_color_combobox.get()
            module = f"{__package__}.app_timer" if __package__ else "app_timer"
            subprocess.Popen(
                [
                    sys.executable,
                    "-m",
                    module,
                    "--file_path",
                    self.csv_path,
                    "--fg_color",
                    color,
                    "--start_index",
                    str(self.start_index),
                    "--hmmss",
                    hmmss,
                ]
            )

    def open_excel(self):
        is_sjis = file_loader.utf8_to_sjis(self.csv_path)
        if is_sjis:
            self.csv_encoding = "shift-jis"
        if IS_DARWIN:
            os.system(f"open -a /Applications/Microsoft/ Excel.app {self.csv_path}")
        else:
            os.system(f'start excel "{self.csv_path}"')

    def export_svg(self):
        init_file_name = os.path.basename(self.csv_path).split(".")[0]
        file_path = filedialog.asksaveasfilename(
            filetypes=[("SVG files", "*.svg")],
            defaultextension=".svg",
            initialfile=init_file_name,
        )
        if not file_path:
            return

        svg_writer.save_as_svg(self.canvas, file_path)

    def undo(self, event):
        backup_dir = "backup"
        # move file backup to tar_path
        backup_file = os.path.join(backup_dir, os.path.basename(self.csv_path))
        if os.path.exists(backup_file):
            shutil.move(backup_file, self.csv_path)
            self.load_file()


def quit(root):
    root.quit()
    root.destroy()


def main():
    bg_color = "#e8e8e8"
    root = ttkthemes.ThemedTk(theme="breeze")
    root.geometry("1400x700+50+50")
    root.configure(background=bg_color)
    root.option_add("*background", bg_color)
    root.option_add("*Canvas.background", bg_color)
    root.option_add("*Text.background", "#fcfcfc")
    root.tk.call("wm", "iconphoto", root._w, tk.PhotoImage(data=icon_data.icon_data))
    s = ttk.Style(root)
    s.configure(".", background=bg_color)
    app = App(root)
    root.protocol("WM_DELETE_WINDOW", lambda: quit(root))
    app.mainloop()


if __name__ == "__main__":
    main()



import tkinter as tk

rect_colors = [
    "#a9a9af",
    "#7c7c83",
    "#d5d5e0",
    "#6a607d",
    "#b0c0d6",
    "#8f8e9e",
    "#a3a3bc",
    "#e0e0fb",
    "#85859f",
    "#c1c1d0",
    "#b0b0ce",
    "#ebebfa",
    "#9b9ba6",
    "#d8d8ef",
]


class Canvas(tk.Canvas):
    fonts = {
        "tiny": {
            "font": ("Helvetica", 8),
            "spacing": 60,
            "height": 8,
        },
        "small": {
            "font": ("Helvetica", 10),
            "spacing": 70,
            "height": 10,
        },
        "normal": {
            "font": ("Helvetica", 12),
            "spacing": 80,
            "height": 12,
        },
    }

    def __init__(self, master, bg="white"):
        super().__init__(master, bg=bg)
        self.mode = "vertical"
        self.rect_width = 15
        self.left_padding = 30
        self.top_padding = 100

    def set_font(self, font):
        self.font = self.fonts[font]["font"]

    def set_scale(self, total_duration):
        self.scale = self.window_size / total_duration

    def set_direction(self, mode):
        if mode not in ["horizontal", "vertical"]:
            raise ValueError("mode must be 'horizontal' or 'vertical'")
        self.mode = mode
        if self.mode == "vertical":
            self.window_size = self.winfo_height() - 40
        else:
            self.window_size = self.winfo_width() - 60

    def draw_start_line(self, start, total_duration):
        self.delete("start_line")
        self.delete("highlight")
        start_sec = start.total_seconds()
        if self.mode == "vertical":
            y_start = start_sec * self.scale + 20
            self._create_start_mark_right(84, y_start, 10, "red")
        else:
            x_start = start_sec * self.scale + self.left_padding
            self._create_start_mark_up(x_start, 200 + self.rect_width, 10, "red")
        self.tag_lower("start_line")
        self.delete("time")

    def _create_start_mark_right(self, x, y, size, color):
        self.create_polygon(
            x - size,
            y - size / 2,
            x,
            y,
            x - size,
            y + size / 2,
            fill=color,
            tag="start_line",
        )

    def _create_start_mark_up(self, x, y, size, color):
        self.create_polygon(
            x - size / 2,
            y + size,
            x,
            y,
            x + size / 2,
            y + size,
            fill=color,
            tag="start_line",
        )

    def create_rect(self, start, duration, color, tag="rect"):
        start_sec = start.total_seconds()
        duration_sec = duration.total_seconds()
        rect_length = duration_sec * self.scale
        if self.mode == "vertical":
            y_start = start_sec * self.scale + 20
            y_end = y_start + rect_length
            self.create_rectangle(
                84,
                y_start,
                84 + self.rect_width,
                y_end,
                fill=color,
                outline="#101010",
                tag=tag,
            )
        else:
            x_start = start_sec * self.scale + self.left_padding
            x_end = x_start + rect_length
            self.create_rectangle(
                x_start,
                self.top_padding,
                x_end,
                self.top_padding + self.rect_width,
                fill=color,
                outline="#101010",
                tag=tag,
            )
        return rect_length

    def create_time(self, start, text):
        start_sec = start.total_seconds()
        if self.mode == "vertical":
            y_start = start_sec * self.scale + 20
            if y_start < 10:
                return
            self.create_text(
                70,
                y_start,
                text=text,
                anchor="e",
                font=self.font,
                tag="time",
            )
        else:
            x_start = start_sec * self.scale + self.left_padding
            self.create_text(
                x_start,
                self.top_padding + 10 + self.rect_width,
                text=text,
                anchor="n",
                font=self.font,
                tag="time",
            )
            self.create_line(
                x_start,
                self.top_padding,
                x_start,
                self.top_padding + 8 + self.rect_width,
                fill="#101010",
                tag="time",
            )

    def create_label(self, start, duration, title_str, time_str, has_error=False):
        start_sec = start.total_seconds()
        duration_sec = duration.total_seconds()
        rect_length = duration_sec * self.scale
        if has_error:
            color = "red"
        elif time_str in ["0:00", "00:00", "0:00:00"]:
            color = "orange"
        else:
            color = "black"
        if self.mode == "vertical":
            y_start = start_sec * self.scale + 20
            y_end = y_start + rect_length
            self.create_text(
                100 + self.rect_width,
                (y_start + y_end) / 2,
                text=f"{title_str} ({time_str})",
                anchor="w",
                font=self.font,
                fill=color,
            )
        else:
            x_start = start_sec * self.scale + self.left_padding
            x_end = x_start + rect_length
            self.create_text(
                (x_start + x_end) / 2,
                self.top_padding - 50,
                text=f"{title_str}\n{time_str}",
                anchor="n",
                font=self.font,
            )
            self.create_line(
                (x_start + x_end) / 2,
                self.top_padding-15,
                (x_start + x_end) / 2,
                self.top_padding,
                fill="#101010",
            )



import tkinter as tk
from tkinter import ttk


class Combobox(ttk.Frame):
    def __init__(self, master, label: str, values: list, width=5, current=0):
        super().__init__(master)
        self.frame = ttk.Frame(master)
        caption = ttk.Label(self.frame, text=label)
        caption.pack(side=tk.LEFT, padx=(0, 1))
        self.combobox = ttk.Combobox(self.frame, state="readonly", width=width)
        self.combobox["values"] = values
        self.combobox.current(current)
        self.combobox.pack(side=tk.LEFT)
        self.current_value = None

    def pack_horizontal(self, anchor=tk.E, padx=0, pady=0, side=tk.LEFT):
        self.frame.pack(side=side, anchor=anchor, padx=padx, pady=pady)

    def set_selected_bind(self, func):
        self.combobox.bind("<<ComboboxSelected>>", func)

    def get(self):
        """Get selected value of the combo box"""
        self.current_value = self.combobox.get()
        return self.combobox.get()

    def get_current_value(self):
        """Get the selected value when get() was called"""
        return self.current_value

    def get_values(self):
        return self.combobox["values"]

    def set(self, value=None):
        """Set value to the combo box"""
        if value is None:
            self.combobox.current(0)
        else:
            self.combobox.set(value)

    def set_values(self, values):
        self.combobox["values"] = values
        self.combobox.current(0)

    def set_state(self, state):
        self.combobox["state"] = state



import argparse
import datetime
import os
import sys
import tkinter as tk
import tomllib
from tkinter import ttk

from . import (
    file_loader,
    gui_ble_button,
    icon_data,
    sound,
    time_format,
    timer_log,
    trigger,
)


class App(ttk.Frame):
    INTERMISSION = "Intermission"

    def __init__(
        self,
        master,
        file_path,
        start_index=0,
        is_hmmss=True,
        sound_file_name="countdown3_orange.wav",
        toml_dict={},
    ):
        super().__init__(master)
        master.title("Timer")
        self.sound_file_name = sound_file_name
        self.hmmss = is_hmmss
        self.master = master

        self.ap = sound.AudioPlayer()
        self.ap.load_audio(sound_file_name)

        self.trigger_device = trigger.Trigger(offset_sec=5)

        # header
        head_frame = ttk.Frame(self.master)
        head_frame.pack(fill=tk.X, pady=10, padx=10)

        title_frame = ttk.Frame(head_frame)
        title_frame.pack(side=tk.LEFT, fill=tk.X)
        clock_frame = ttk.Frame(head_frame)
        clock_frame.pack(side=tk.RIGHT, fill=tk.X)

        self.title_label = ttk.Label(title_frame, text="", font=("Helvetica", 28))
        self.title_label.pack(anchor=tk.W)

        self.main_clock_label = ttk.Label(clock_frame, font=("Helvetica", 12))
        self.main_clock_label.pack(anchor=tk.E, pady=1)
        self.count_up_label = ttk.Label(clock_frame, font=("Helvetica", 18))
        self.count_up_label.pack(anchor=tk.E)

        # stage
        center_frame = ttk.Frame(self.master)
        center_frame.pack(fill=tk.BOTH, expand=True)
        self.current_stage_label = ttk.Label(center_frame, style="Small.TLabel")
        self.current_stage_label.grid(row=0, column=0, sticky=tk.S)
        self.current_instruction_label = ttk.Label(center_frame, font=("Helvetica", 18))
        self.current_instruction_label.grid(row=1, column=0, sticky=tk.N)
        center_frame.columnconfigure(0, weight=1)
        center_frame.rowconfigure(0, weight=1)
        center_frame.rowconfigure(1, weight=1)

        # footer
        buttons_frame = ttk.Frame(self.master)
        buttons_frame.pack(pady=(10, 5), side=tk.BOTTOM, fill=tk.BOTH, anchor=tk.S)

        # progress
        progress_frame = ttk.Frame(buttons_frame)
        progress_frame.pack(pady=10, fill=tk.X)
        window_width = self.winfo_screenwidth()
        self.progress_bar = ttk.Progressbar(
            progress_frame, orient=tk.HORIZONTAL, length=window_width
        )
        self.progress_bar.pack()
        self.progress_bar["value"] = 0
        self.progress_bar["maximum"] = 100

        # next stage
        next_frame = ttk.Frame(buttons_frame)
        next_frame.pack(padx=10, pady=(0, 15), fill=tk.X)
        self.next_stage_label = ttk.Label(next_frame, font=("Helvetica", 18))
        self.next_stage_label.pack(side=tk.LEFT, anchor=tk.E)

        self.remaining_time_label = ttk.Label(next_frame, font=("Helvetica", 18))
        self.remaining_time_label.pack(padx=(15, 0), side=tk.LEFT, anchor=tk.E)

        self.start_btn = ttk.Button(buttons_frame, text="Start", command=self.start)
        self.start_btn.pack(padx=12, side=tk.LEFT)
        self.start_btn["state"] = "disabled"

        self.sound_test_btn = ttk.Button(
            buttons_frame,
            text="Sound test",
            command=lambda: self.ap.play_sound(self.sound_file_name),
        )
        self.sound_test_btn.pack(padx=12, side=tk.LEFT)

        self.reset_btn = ttk.Button(buttons_frame, text="Reset", command=self.reset_all)
        self.reset_btn.pack(padx=12, side=tk.LEFT)
        self.reset_btn["state"] = "disabled"

        switch_label_size_button = ttk.Button(
            buttons_frame, text="Label size", command=self.switch_label_size
        )
        switch_label_size_button.pack(padx=12, side=tk.LEFT)

        skip_btn = ttk.Button(buttons_frame, text="Skip", command=self.skip)
        skip_btn.pack(padx=12, side=tk.LEFT)

        ble_names = toml_dict.get("ble_names", [])
        if len(ble_names) > 0:
            self.enable_ble = True
            self.stop_delay_sec = toml_dict.get("stop_delay_sec", 1)
            self.ble_manager = gui_ble_button.BleButtonManager(
                buttons_frame,
                self.master,
                self.trigger_device,
                ble_names,
                self.stop_delay_sec,
            )
        else:
            self.enable_ble = False

        # close event
        self.master.protocol("WM_DELETE_WINDOW", self._on_closing)

        self.stage_list = []
        self.now_stage = 0
        self.is_running = False
        self.update_clock()
        self.ring_done = False
        self.is_skip = False
        self.disp_time = datetime.timedelta(seconds=0)

        print(f"CSV file path: {file_path}")
        self.csv_path = file_path
        #        self.tlog = timer_log.TimerLog(self.csv_path)

        self.bids_log = timer_log.BIDSLog(self.csv_path)

        self.load_file(start_index)

    def update_clock(self):
        now = datetime.datetime.now()
        self.main_clock_label.config(text=now.strftime("%Y-%m-%d %H:%M:%S"))
        self.after(100, self.update_clock)

        # If the timer is stopped, reset the timer
        if self.is_running is False:
            self.count_up_label.config(text="00:00:00")
            self.now_stage = 0
            self.progress_bar.config(value=0)
            self.reset_time = datetime.datetime.now()
            self.total_skip_time = datetime.timedelta(seconds=0)
            self.is_skip = False
            if self.enable_ble:
                self.ble_manager.update_ble_status()
            return
        # cnt_up: internal time counter
        # self.disp_time: time used for display and log
        cnt_up = now - self.reset_time
        self.disp_time = cnt_up - self.total_skip_time
        self.count_up_label.config(
            text=time_format.timedelta_to_str(self.disp_time, self.hmmss)
        )

        current_stage = self.stage_list[self.now_stage]
        current_start_dt = current_stage["start_dt"]
        current_end_dt = current_stage["end_dt"]
        # stage change if end time is reached
        if cnt_up > current_end_dt:
            self.now_stage += 1
            if self.now_stage < len(self.stage_list):
                print(
                    f"stage change {self.now_stage} (/{len(self.stage_list)}) {cnt_up}"
                )
                current_stage = self.stage_list[self.now_stage]
                current_start_dt = current_stage["start_dt"]
                current_end_dt = current_stage["end_dt"]
                self.bids_log.add_task_log()
                if cnt_up < current_start_dt:
                    msg = f"{self.INTERMISSION}({current_stage['start_dt']}-{current_stage['end_dt']})"
                    self.bids_log.set_task_log(self.INTERMISSION)
                else:
                    msg = f"{current_stage['title']}({current_stage['start_dt']}-{current_stage['end_dt']})"
                    self.bids_log.set_task_log(current_stage["title"])
        #                self.tlog.add_log(self.disp_time, msg)

        # If the last stage is reached, stop the timer
        if self.now_stage >= len(self.stage_list):
            self.current_stage_label.config(text="End")
            self.current_instruction_label.config(text="")
            self.next_stage_label.config(text="---")
            self.remaining_time_label.config(text="")
            self.is_running = False
            self.bids_log.add_task_log()
            self.bids_log.add_control_log("session_end")
            self.trigger_device.trigger_out("End")
            return

        self.current_stage_label.config(text=current_stage["title"])
        self.title_label.config(text=current_stage["member"])
        self.update_next_stage_label(end_text="End")
        self.update_instruction_label(current_stage, current_start_dt, current_end_dt)

        remaining_dt = self.calc_remaining_time_next(cnt_up, current_end_dt)
        self.sound(remaining_dt, offset_sec=3)
        # BLE
        if self.enable_ble:
            self.trigger(self.now_stage, remaining_dt)
            self.ble_manager.update_ble_status()

        self.update_remaining_time_label(remaining_dt)
        self.update_progress_bar(cnt_up, current_end_dt)
        self.update_skip(remaining_dt, offset_sec=4)

    def update_instruction_label(self, current_stage, start_dt, end_dt):
        # Display instruction if it exists, otherwise display start and end time
        if current_stage["instruction"]:
            self.current_instruction_label.config(text=current_stage["instruction"])
        else:
            start_str = time_format.timedelta_to_str(start_dt, self.hmmss)
            end_str = time_format.timedelta_to_str(end_dt, self.hmmss)
            self.current_instruction_label.config(text=f"{start_str} - {end_str}")

    def calc_remaining_time_next(self, cnt_up, next_dt):
        if next_dt < cnt_up:
            remaining_dt = datetime.timedelta(seconds=0)
        else:
            remaining_dt = next_dt - cnt_up + datetime.timedelta(seconds=1)
        return remaining_dt

    def sound(self, remaining_dt, offset_sec=3):
        if remaining_dt.seconds == offset_sec and not self.ring_done:
            self.ap.play_sound(self.sound_file_name)
            self.ring_done = True
        if remaining_dt.seconds < offset_sec:
            self.ring_done = False

    def trigger(self, now_stage, remaining_dt):
        is_start_trigger = False
        next_stage = now_stage + 1
        if next_stage > len(self.stage_list):
            return

        if next_stage == len(self.stage_list):
            next_stage_instruction = "End"
        else:
            next_stage_instruction = self.stage_list[next_stage]["instruction"]
        current_stage_instruction = self.stage_list[now_stage]["instruction"]
        # Trigger in before X seconds of next stage
        if remaining_dt < datetime.timedelta(seconds=self.trigger_device.offset_sec):
            is_start_trigger = self.trigger_device.trigger_in(next_stage_instruction)
        else:
            # for 1st stage
            is_start_trigger = self.trigger_device.trigger_in(current_stage_instruction)
            # Trigger out when leaving current stage
            is_stop_trigger = self.trigger_device.trigger_out(current_stage_instruction)

        if is_start_trigger:
            #            self.tlog.add_log(self.disp_time, "recording start triggered")
            self.bids_log.add_control_log("video_record_start")

    def update_progress_bar(self, cnt_up, next_dt):
        duration_dt = self.stage_list[self.now_stage]["duration"]
        progress = cnt_up - next_dt
        val = 100 + progress / duration_dt * 100
        self.progress_bar.config(value=val)

    def update_next_stage_label(self, end_text="End"):
        if self.now_stage + 1 < len(self.stage_list):
            self.next_stage_label["text"] = self.stage_list[self.now_stage + 1]["title"]
        else:
            self.next_stage_label["text"] = end_text

    def update_remaining_time_label(self, remaining_dt):
        if remaining_dt <= datetime.timedelta(seconds=0):
            self.remaining_time_label.config(text="")
        else:
            self.remaining_time_label.config(
                text=time_format.timedelta_to_str(remaining_dt, self.hmmss)
            )

    def update_skip(self, remaining_dt, offset_sec=4):
        if self.is_skip:
            # prevent over skip
            if remaining_dt > datetime.timedelta(seconds=offset_sec):
                skip_time = remaining_dt - datetime.timedelta(seconds=offset_sec)
                self.reset_time -= skip_time
                self.total_skip_time += skip_time
                #                self.tlog.skip_log(self.disp_time)
                self.bids_log.add_control_log("task_skip")
            self.is_skip = False

    def reset_all(self):
        self.is_running = False
        #        self.tlog.reset_log(self.disp_time)
        self.start_btn.config(state="normal")
        self.sound_test_btn.config(state="normal")
        self.reset_btn.config(state="disabled")
        self.current_stage_label.config(text="")
        self.current_instruction_label.config(text="")
        self.next_stage_label.config(text=self.stage_list[0]["title"])
        self.remaining_time_label.config(text="")

        # BLE stop
        self.trigger_device.trigger_out("")

        # for close log
        self.disp_time = datetime.timedelta(seconds=0)

    def start(self):
        #        self.tlog.start_log()
        self.bids_log.mark_start_time(datetime.datetime.now())
        self.start_btn.config(state="disabled")
        self.sound_test_btn.config(state="disabled")
        self.reset_btn.config(state="normal")
        if self.enable_ble:
            self.ble_manager.set_disabled()
        self.is_running = True

        # initial event log
        current_stage = self.stage_list[0]
        msg = f"{current_stage['title']}({current_stage['start_dt']}-{current_stage['end_dt']})"
        #        self.tlog.add_log(self.disp_time, msg)
        self.bids_log.set_task_log(current_stage["title"])

    def skip(self):
        self.is_skip = True

    def switch_label_size(self):
        if self.current_stage_label["style"] == "Small.TLabel":
            self.current_stage_label["style"] = "Large.TLabel"
        elif self.current_stage_label["style"] == "Large.TLabel":
            self.current_stage_label["style"] = "Tiny.TLabel"
        else:
            self.current_stage_label["style"] = "Small.TLabel"

    def load_file(self, start_index):
        fl = file_loader.FileLoader(self.INTERMISSION)
        fl.load_file_for_timer(start_index, self.csv_path)

        self.stage_list = fl.get_stage_list()

        self.title_label.config(text=self.stage_list[0]["member"])
        self.next_stage_label.config(text=self.stage_list[0]["title"])
        self.now_stage = 0
        self.is_running = False
        self.start_btn["state"] = "normal"

        for i, stage in enumerate(self.stage_list):
            print(
                f"{i}: {stage['title']}({stage['start_dt']}-{stage['end_dt']}) {stage['instruction']}"
            )

    def _on_closing(self):
        self.trigger_device.trigger_out("")
        #        self.tlog.close_log(self.disp_time)
        self.master.quit()
        self.master.destroy()


def main(
    file_path=None, fg_color: str = "orange", start_index: int = 0, hmmss: str = "hmmss"
):
    bg_color = "#202020"
    color_and_sound = {
        "orange": "countdown3_orange.wav",
        "cyan": "countdown3_cyan.wav",
        "lightgreen": "countdown3_lightgreen.wav",
    }

    root = tk.Tk()
    root.geometry("900x420+0+0")
    root.configure(background=bg_color)
    root.tk.call("wm", "iconphoto", root._w, tk.PhotoImage(data=icon_data.icon_data))
    s = ttk.Style(root)
    s.theme_use("default")
    s.configure("TFrame", background=bg_color)
    s.configure("TLabel", foreground=fg_color, background=bg_color)
    s.configure(
        "Tiny.TLabel", foreground=fg_color, background=bg_color, font=("Helvetica", 36)
    )
    s.configure(
        "Small.TLabel", foreground=fg_color, background=bg_color, font=("Helvetica", 48)
    )
    s.configure(
        "Large.TLabel", foreground=fg_color, background=bg_color, font=("Helvetica", 64)
    )
    s.configure("TButton", foreground=fg_color, background=bg_color, relief="flat")
    s.map("TButton", background=[("active", "#203030")], relief=[("active", "flat")])
    s.configure(
        "Horizontal.TProgressbar",
        troughcolor=bg_color,
        troughrelief="flat",
        background=fg_color,
        borderwidth=0,
        thickness=5,
    )
    s.map(
        "Horizontal.TProgressbar",
        troughcolor=[("active", bg_color), ("!active", bg_color)],
    )

    if hmmss == "hmmss":
        is_hmmss = True
    else:
        is_hmmss = False

    # BLE
    if getattr(sys, "frozen", False):
        current_dir = os.path.dirname(sys.executable)
    else:
        current_dir = os.path.dirname(__file__)
    ble_file_name = os.path.join(current_dir, "ble_devices.toml")
    with open(ble_file_name, "rb") as f:
        toml = tomllib.load(f)
    toml_dict = toml.get(fg_color, {})

    app = App(
        root,
        file_path,
        start_index,
        is_hmmss,
        sound_file_name=color_and_sound[fg_color],
        toml_dict=toml_dict,
    )
    app.mainloop()


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Timeline-kun Timer: manage timed tasks with GUI."
    )
    parser.add_argument(
        "--file_path",
        type=str,
        required=True,
        help="Path to the CSV file containing task definitions (required).",
    )
    parser.add_argument(
        "--fg_color",
        type=str,
        default="orange",
        choices=["orange", "cyan", "lightgreen"],
        help="Foreground color for text and theme (default: orange).",
    )
    parser.add_argument(
        "--start_index",
        type=int,
        default=0,
        help="Index of the task to start from (default: 0). Must be an integer.",
    )
    parser.add_argument(
        "--hmmss",
        type=str,
        default="hmmss",
        choices=["hmmss", "mmss"],
        help="Display format for time (default: hmmss). 'hmmss' for 1:23:45, 'mmss' for 83:45.",
    )
    return parser


def cli(argv=None):
    parser = build_parser()
    args = parser.parse_args(argv)
    return main(args.file_path, args.fg_color, args.start_index, args.hmmss)


if __name__ == "__main__":
    cli()



# ble_control.py
import asyncio
import queue
import threading
from dataclasses import dataclass, field
from typing import Any, Awaitable, Callable, Dict, List, Optional, Tuple

from bleak import BleakClient, BleakScanner

Result = Tuple[str, int, str]
CommandHandler = Callable[[Optional[Any]], Awaitable[Result]]


@dataclass(frozen=True)
class BleData:
    COMMAND_UUID: str = "b5f90072-aa8d-11e3-9046-0002a5d5c51b"
    SETTING_UUID: str = "b5f90074-aa8d-11e3-9046-0002a5d5c51b"
    RESPONSE_UUID: str = "b5f90075-aa8d-11e3-9046-0002a5d5c51b"

    START_RECORDING: bytearray = field(
        default_factory=lambda: bytearray([0x03, 0x01, 0x01, 0x01])
    )
    STOP_RECORDING: bytearray = field(
        default_factory=lambda: bytearray([0x03, 0x01, 0x01, 0x00])
    )
    KEEP_ALIVE: bytearray = field(
        default_factory=lambda: bytearray([0x03, 0x5B, 0x01, 0x42])
    )

    KEEP_ALIVE_ID: int = 0x5B
    OK: int = 0x00
    RESP_PREFIX: int = 0x02


BLE = BleData()


@dataclass
class BLEMessage:
    raw: bytes
    msg_type: int
    msg_id: int
    status: int
    payload: bytes

    @classmethod
    def parse(cls, data: bytearray) -> Optional["BLEMessage"]:
        b = bytes(data)
        if len(b) < 3:
            return None
        return cls(raw=b, msg_type=b[0], msg_id=b[1], status=b[2], payload=b[3:])


NotifyHandler = Callable[[int, bytearray], Awaitable[None]]


class NotifyCenter:
    """デバイス単位の通知集約"""

    def __init__(self) -> None:
        self._queues: Dict[int, asyncio.Queue[BLEMessage]] = {}
        self._handler: Optional[NotifyHandler] = None

    def _queue_for(self, msg_id: int) -> asyncio.Queue[BLEMessage]:
        q = self._queues.get(msg_id)
        if q is None:
            q = asyncio.Queue(maxsize=16)
            self._queues[msg_id] = q
        return q

    def get_handler(self) -> NotifyHandler:
        if self._handler is None:

            async def handler(sender: int, data: bytearray) -> None:
                msg = BLEMessage.parse(data)
                if msg is None:
                    return
                await self._safe_put(self._queue_for(msg.msg_id), msg)

            self._handler = handler
        return self._handler

    async def _safe_put(self, q: asyncio.Queue, item: BLEMessage) -> None:
        try:
            q.put_nowait(item)
        except asyncio.QueueFull:
            try:
                _ = q.get_nowait()
            except asyncio.QueueEmpty:
                pass
            await q.put(item)

    async def wait_for(self, msg_id: int, timeout: float) -> Optional[BLEMessage]:
        q = self._queue_for(msg_id)
        try:
            return await asyncio.wait_for(q.get(), timeout=timeout)
        except asyncio.TimeoutError:
            return None

    def clear(self, msg_id: Optional[int] = None) -> None:
        if msg_id is None:
            for q in self._queues.values():
                self._drain(q)
        else:
            q = self._queues.get(msg_id)
            if q:
                self._drain(q)

    @staticmethod
    def _drain(q: asyncio.Queue) -> None:
        try:
            while True:
                q.get_nowait()
        except asyncio.QueueEmpty:
            return


class DeviceSession:
    def __init__(self, address: str) -> None:
        self.address = address
        self.client = BleakClient(address)
        self.center = NotifyCenter()
        self.last_alive: bool = False

    async def connect_and_listen(self) -> bool:
        if not self.client.is_connected:
            print(f"connecting {self.address}")
            await self.client.connect()
        await self.client.start_notify(BLE.RESPONSE_UUID, self.center.get_handler())
        return True

    async def ensure_connected(self) -> bool:
        if self.client.is_connected:
            return True
        try:
            await self.connect_and_listen()
            print(f"reconnected {self.address}")
            return True
        except Exception as e:
            print(f"reconnect failed {self.address} {e}")
            return False

    async def write_command(self, uuid: str, payload: bytearray) -> bool:
        if not await self.ensure_connected():
            return False
        try:
            await self.client.write_gatt_char(uuid, payload, response=True)
            return True
        except Exception as e:
            print(f"write failed {self.address} {e}")
            return False

    async def keep_alive_roundtrip(self, timeout: float = 3.0) -> bool:
        self.center.clear(BLE.KEEP_ALIVE_ID)
        if not await self.write_command(BLE.SETTING_UUID, BLE.KEEP_ALIVE):
            self.last_alive = False
            return False
        msg = await self.center.wait_for(BLE.KEEP_ALIVE_ID, timeout=timeout)
        if msg and msg.status != BLE.OK:
            print(msg)
        ok = bool(msg and msg.msg_type == BLE.RESP_PREFIX and msg.status == BLE.OK)
        self.last_alive = ok
        return ok


class BleManager:
    def __init__(
        self, target_device_names: List[str], keep_alive_sec: float = 3.0
    ) -> None:
        self.target_device_names = target_device_names
        self.sessions: List[DeviceSession] = []
        self.keep_alive_sec = keep_alive_sec
        self._keep_task: Optional[asyncio.Task] = None
        self._running = False
        self.is_recording = False

    async def discover_and_connect(self) -> int:
        devices = await BleakScanner.discover(timeout=10)
        names = set(n for n in self.target_device_names if n)
        addrs: List[str] = []
        for d in devices:
            if d.name and d.name in names:
                print(f"found {d.name} {d.address}")
                addrs.append(d.address)

        self.sessions = [DeviceSession(addr) for addr in addrs]
        ok = 0
        for s in self.sessions:
            try:
                await s.connect_and_listen()
                ok += 1
            except Exception as e:
                print(f"connect failed {s.address} {e}")

        if ok:
            self.start_keep_alive_loop()
        return ok

    def start_keep_alive_loop(self) -> None:
        if self._keep_task is None or self._keep_task.done():
            self._running = True
            self._keep_task = asyncio.create_task(self._keep_loop())

    def stop_keep_alive_loop(self) -> None:
        self._running = False
        if self._keep_task:
            self._keep_task.cancel()

    async def _keep_loop(self) -> None:
        while self._running:
            await self.send_keep_alive_all()
            try:
                await asyncio.sleep(self.keep_alive_sec)
            except asyncio.CancelledError:
                break

    async def send_keep_alive_all(self) -> int:
        ok_cnt = 0
        for s in list(self.sessions):
            alive = await s.keep_alive_roundtrip(timeout=3.0)
            if not alive:
                # 再接続のチャンスをここで与える
                if await s.ensure_connected():
                    alive = await s.keep_alive_roundtrip(timeout=3.0)
            if alive:
                ok_cnt += 1
        #        print(f"keep alive {ok} of {len(self.sessions)}")
        if ok_cnt < len(self.sessions):
            print(f"[!] keep alive ok {ok_cnt} of {len(self.sessions)}")
        return ok_cnt

    async def start_recording_all(self) -> int:
        cnt = 0
        for s in self.sessions:
            if await s.write_command(BLE.COMMAND_UUID, BLE.START_RECORDING):
                cnt += 1
        self.is_recording = cnt > 0
        return cnt

    async def stop_recording_all(self) -> int:
        cnt = 0
        for s in self.sessions:
            if await s.write_command(BLE.COMMAND_UUID, BLE.STOP_RECORDING):
                cnt += 1
        self.is_recording = False if cnt > 0 else self.is_recording
        return cnt


class BleThread:
    """スレッドはキュー処理だけに限定"""

    def __init__(self) -> None:
        self.thread: Optional[threading.Thread] = None
        self.loop: Optional[asyncio.AbstractEventLoop] = None
        self.running = False
        self.command_q: "queue.Queue[Tuple[str, Optional[Any]]]" = queue.Queue()
        self.result_q: "queue.Queue[Result]" = queue.Queue()
        self.manager: Optional[BleManager] = None
        self.target_device_names: List[str] = [""]
        self._handlers: Dict[str, CommandHandler] = {
            "connect": self._h_connect,
            "record_start": self._h_start,
            "record_stop": self._h_stop,
            "status": self._h_status,
        }

    def set_target_device_names(self, names: List[str]) -> None:
        self.target_device_names = names

    def start(self) -> None:
        if self.thread and self.thread.is_alive():
            return
        self.running = True
        self.thread = threading.Thread(target=self._run, daemon=True)
        self.thread.start()

    def stop(self) -> None:
        if not self.running:
            return
        self.running = False
        self.command_q.put(("__stop__", None))
        if self.thread and self.thread.is_alive():
            self.thread.join()

    def execute_command(
        self, command: str, data: Optional[Any] = None, timeout: int = 30
    ) -> Result:
        if not self.running:
            return (command, 0, "thread not running")
        self.command_q.put((command, data))
        try:
            return self.result_q.get(timeout=timeout)
        except queue.Empty:
            return (command, 0, "timeout")

    def _run(self) -> None:
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        try:
            self.manager = BleManager(self.target_device_names, keep_alive_sec=3.0)
            self.loop.run_until_complete(self._loop())
        finally:
            if self.loop:
                self.loop.close()

    async def _loop(self) -> None:
        assert self.manager is not None
        while self.running:
            try:
                cmd, data = self.command_q.get_nowait()
            except queue.Empty:
                await asyncio.sleep(0.05)
                continue

            if cmd == "__stop__":
                self.manager.stop_keep_alive_loop()
                break

            handler = self._handlers.get(cmd)
            if handler is None:
                self.result_q.put((cmd, 0, "unknown"))
                continue
            try:
                res = await handler(data)
            except Exception as e:
                res = (cmd, 0, f"error {e}")
            self.result_q.put(res)

    # handlers
    async def _h_connect(self, _: Optional[Any]) -> Result:
        assert self.manager is not None
        self.manager.target_device_names = self.target_device_names
        ok = await self.manager.discover_and_connect()
        return ("connect", ok, "success" if ok else "failed")

    async def _h_start(self, _: Optional[Any]) -> Result:
        assert self.manager is not None
        ok = await self.manager.start_recording_all()
        return ("record_start", ok, "success" if ok else "failed")

    async def _h_stop(self, _: Optional[Any]) -> Result:
        assert self.manager is not None
        ok = await self.manager.stop_recording_all()
        return ("record_stop", ok, "success" if ok else "failed")

    async def _h_status(self, _: Optional[Any]) -> Result:
        assert self.manager is not None
        total = len(self.manager.sessions)
        alive_cnt = sum(1 for s in self.manager.sessions if s.last_alive)
        state = "recording" if self.manager.is_recording else "idle"
        return ("status", alive_cnt, f"{alive_cnt}/{total} {state}")



import os
from datetime import datetime

from . import time_format


class TimerLog:
    def __init__(self, csv_file_path):
        tar_dir = os.path.dirname(csv_file_path)
        tar_name = os.path.basename(csv_file_path).split(".")[0]
        today = datetime.now().strftime("%Y-%m-%d")
        self.file_path = os.path.join(tar_dir, f"log_{today}_{tar_name}.csv")
        if os.path.exists(self.file_path) is False:
            with open(self.file_path, "w") as f:
                f.write("datetime,displaytime,message\n")

    def add_log(self, display_time, message):
        dt_str = time_format.timedelta_to_str(display_time)
        self._write_log(dt_str, message)

    def start_log(self):
        print("start")
        self._write_log("0:00:00", "====== start ======")

    def skip_log(self, display_time):
        print("skip")
        dt_str = time_format.timedelta_to_str(display_time)
        self._write_log(dt_str, "skip")

    def reset_log(self, display_time):
        print("reset")
        dt_str = time_format.timedelta_to_str(display_time)
        self._write_log(dt_str, "reset")

    def close_log(self, display_time):
        print("close")
        dt_str = time_format.timedelta_to_str(display_time)
        self._write_log(dt_str, "close")

    def end_log(self, display_time):
        print("end")
        dt_str = time_format.timedelta_to_str(display_time)
        self._write_log(dt_str, "end")

    def _write_log(self, display_time, message):
        now_str = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        with open(self.file_path, "a") as f:
            f.write(f"{now_str},{display_time},{message}\n")


class BIDSLog:
    CONTROL_LOGS = [
        "video_record_start",
        "task_skip",
        "session_end",
    ]

    def __init__(self, csv_file_path):
        # BIDS events.tsv format
        tar_dir = os.path.dirname(csv_file_path)
        tar_name = os.path.basename(csv_file_path).split(".")[0]
        self.events_path = os.path.join(tar_dir, f"{tar_name}_")

        # scans.tsv file
        self.scans_path = os.path.join(tar_dir, f"{tar_name}_scans.tsv")
        if os.path.exists(self.scans_path) is False:
            with open(self.scans_path, "w") as f:
                f.write("filename\tacq_time\n")
        self.task_name = None
        self.task_start_dt = None

    def _write_log(self, onset, duration, trial_type):
        with open(self.file_path, "a") as f:
            f.write(f"{onset:<.1f}\t{duration:<.1f}\t{trial_type}\n")

    def _write_scans_log(self, filename, acq_time):
        with open(self.scans_path, "a") as f:
            f.write(f"{filename}\t{acq_time}\n")

    def mark_start_time(self, start_time):
        self.session_start_time = start_time

        # events.tsv file
        for i in range(100):
            if os.path.exists(self.events_path + f"{i:0>2}_events.tsv") is False:
                self.file_path = self.events_path + f"{i:0>2}_events.tsv"
                break
        with open(self.file_path, "w") as f:
            f.write("onset\tduration\ttrial_type\n")

        with open(self.scans_path, "a") as f:
            f.write(
                f"{os.path.basename(self.file_path)}\t{datetime.now().strftime('%Y-%m-%dT%H:%M:%S')}\n"
            )

    def add_control_log(self, trial_type):
        if trial_type not in self.CONTROL_LOGS:
            raise ValueError(f"Invalid control log type: {trial_type}")
        onset = (datetime.now() - self.session_start_time).total_seconds()
        duration = 0.0
        self._write_log(onset, duration, trial_type)
        # for MP4 file in Cameras
        if trial_type == "video_record_start":
            self._write_scans_log(
                "Unknown", datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            )

    def set_task_log(self, task_name: str):
        self.task_start_time = datetime.now()
        self.task_name = task_name

    def add_task_log(self):
        onset = (self.task_start_time - self.session_start_time).total_seconds()
        duration = (datetime.now() - self.task_start_time).total_seconds()
        self._write_log(onset, duration, self.task_name)



import datetime
import re
import sys
import tkinter as tk
from tkinter import ttk

from . import time_format, timetable_to_csv

IS_DARWIN = sys.platform.startswith("darwin")


class Tree(ttk.Frame):
    """
    End column is not displayed in the tree.
    If you set the end time, csv file will have the end column.
    """

    def __init__(self, master, columns: list):
        super().__init__(master)
        cols = [col["name"] for col in columns]
        self.tree = ttk.Treeview(
            self, columns=cols, show="headings", selectmode="browse"
        )
        for column in columns:
            self.tree.heading(column["name"], text=column["name"])
            self.tree.column(column["name"], width=column["width"])
        self.tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        self.tree["displaycolumns"] = (
            "title",
            "member",
            "start",
            "duration",
            "fixed",
            "instruction",
        )
        if not IS_DARWIN:
            self.tree.bind("<Button-3>", self._right_click_tree)
        else:
            self.tree.bind("<Button-2>", self._right_click_tree)

        self.menu = tk.Menu(self, tearoff=0)
        self.stage_list = []
        self.write_encoding = "utf-8"

    def set_stages(self, stages):
        self.tree.delete(*self.tree.get_children())
        for i, stage in enumerate(stages):
            self.tree.insert(
                "",
                "end",
                values=[
                    stage["title"],
                    stage["member"],
                    time_format.seconds_to_time_str(stage["start_sec"]),
                    time_format.seconds_to_time_str(stage["end_sec"]),
                    time_format.seconds_to_time_str(stage["duration_sec"]),
                    stage["fixed"],
                    stage["instruction"],
                ],
            )
        self.stage_list = stages

    def add_menu(self, label, command):
        self.menu.add_command(label=label, command=command)

    def set_write_encoding(self, encoding):
        self.write_encoding = encoding

    def get_selected_index(self):
        selected = self.tree.selection()
        if len(selected) == 0:
            return None
        selected = selected[0]
        return self.tree.index(selected)

    def _right_click_tree(self, event):
        selected = self.tree.selection()
        if len(selected) == 0:
            return
        self.menu.post(event.x_root, event.y_root)

    def check_csv_file_locked(self, file_path):
        if timetable_to_csv.check_file_locked(file_path) is True:
            return True
        return False

    def edit(self, prev_end_sec):
        """Only one row can be edited at once"""
        selected = self.tree.selection()
        if len(selected) == 0 or len(selected) > 1:
            return False
        default_row = self.tree.item(selected)["values"]

        x = self.winfo_rootx()
        y = self.winfo_rooty()
        dialog = TimelineTreeDialog(self, x, y)
        dialog.set_row_contents(default_row)

        idx = self.get_selected_index()
        start = time_format.timedelta_to_str(self.stage_list[idx]["start_dt"])
        end = time_format.timedelta_to_str(self.stage_list[idx]["end_dt"])
        dialog.set_current_time_range(start, end)
        dialog.set_prev_sec(prev_end_sec)
        self.wait_window(dialog.dialog)

        new_title = dialog.selected_title
        new_member = dialog.selected_member
        if new_title is None:
            return False
        for item in selected:
            values = self.tree.item(item)["values"]
            if new_title != "":
                values[0] = new_title
            if new_member != "":
                values[1] = new_member
            if dialog.selected_fixed != "":
                values[5] = dialog.selected_fixed
            values[2] = dialog.selected_start
            values[3] = dialog.selected_end
            values[4] = dialog.selected_duration
            if dialog.selected_instruction != "":
                values[6] = dialog.selected_instruction
            elif dialog.selected_instruction == " ":
                values[6] = ""

            self.tree.item(item, values=values)
        return True

    def insert(self):
        selected = self.tree.selection()
        if len(selected) != 1:
            return False

        item = selected[0]
        idx = self.tree.index(item)
        if idx == 0:
            start_dt = datetime.timedelta(seconds=0)
        else:
            start_dt = self.stage_list[idx - 1]["end_dt"]

        new_row = {
            "title": "New event",
            "member": "",
            "start_dt": start_dt,
            "end_dt": start_dt,
            "duration": datetime.timedelta(seconds=60),
            "start_sec": "",
            "end_sec": "",
            "duration_sec": "1:00",
            "fixed": "duration",
            "instruction": "",
        }
        item = selected[0]
        idx = self.tree.index(item)
        self.stage_list.insert(idx, new_row)
        self.set_stages(self.stage_list)
        return True

    def remove(self):
        """Only one row can be removed at once"""
        selected = self.tree.selection()
        if len(selected) != 1:
            return False

        item = selected[0]

        # update start and end time below all removed item
        idx = self.tree.index(item)
        self.stage_list.pop(idx)
        self.set_stages(self.stage_list)
        return True

    def tree_to_csv_file(self, file_path):
        end_col_num = 3
        timetable_to_csv.move_to_backup_folder(file_path)
        ends = [
            self.tree.item(item)["values"][end_col_num]
            for item in self.tree.get_children()
        ]

        if ends.count("") == len(ends):
            header = "title,member,start,duration,fixed,instruction"
            no_end = True
        else:
            header = "title,member,start,end,duration,fixed,instruction"
            no_end = False

        with open(file_path, "w", encoding=self.write_encoding) as f:
            f.write(header + "\n")
            for item in self.tree.get_children():
                values = self.tree.item(item)["values"]
                if no_end:
                    values.pop(end_col_num)
                line = ",".join(str(val) for val in values)
                f.write(line + "\n")


class TimelineTreeDialog(tk.Frame):
    def __init__(self, master, x, y):
        super().__init__(master)
        dialog = tk.Toplevel(master)
        dialog.focus_set()
        dialog.title("Timeline row")
        dialog.geometry(f"+{x + 500}+{y}")
        dialog.grab_set()

        tar_frame = ttk.Frame(dialog)
        tar_frame.pack(side=tk.TOP, padx=20, pady=(20, 0))

        title_frame = ttk.Frame(tar_frame)
        title_frame.pack(pady=5, side=tk.TOP, anchor=tk.W)
        title_label = ttk.Label(title_frame, text="Title:")
        title_label.pack(side=tk.LEFT)
        self.title_entry = ttk.Entry(title_frame, width=27)
        self.title_entry.pack(padx=(0, 30), side=tk.LEFT)

        member_label = ttk.Label(title_frame, text="Member:")
        member_label.pack(side=tk.LEFT)
        self.member_entry = ttk.Entry(title_frame, width=27)
        self.member_entry.pack(side=tk.LEFT)

        instruction_frame = ttk.Frame(tar_frame)
        instruction_frame.pack(pady=5, side=tk.TOP, anchor=tk.W)
        instruction_label = ttk.Label(instruction_frame, text="Instruction:")
        instruction_label.pack(padx=(0, 3), side=tk.LEFT)
        self.instructions_entry = ttk.Entry(instruction_frame, width=62)
        self.instructions_entry.pack(side=tk.LEFT)

        self.current_time_range_label = TimeRangeLabel(tar_frame, "Current time range:")
        self.current_time_range_label.pack(pady=5, side=tk.TOP, anchor=tk.W)
        self.msg_label = ttk.Label(tar_frame, text="")
        self.msg_label.pack(pady=5, side=tk.TOP, anchor=tk.W)

        time_frame = ttk.Frame(tar_frame)
        time_frame.pack(pady=5, side=tk.TOP, anchor=tk.W)

        fixed_frame = ttk.Frame(time_frame)
        fixed_frame.pack(padx=(0, 20), pady=5, anchor=tk.W, side=tk.LEFT)
        fixed_label = ttk.Label(fixed_frame, text="Fixed:")
        fixed_label.pack(side=tk.LEFT)
        self.fixed_combobox = ttk.Combobox(
            fixed_frame, values=["start", "duration", ""], state="readonly", width=10
        )
        self.fixed_combobox.pack(side=tk.LEFT)
        self.fixed_combobox.bind("<<ComboboxSelected>>", self.on_fixed_selected)

        self.start_entry = TimeEntry(time_frame, "Start:")
        self.start_entry.pack(padx=(0, 20), side=tk.LEFT)

        self.duration_entry = TimeEntry(time_frame, "Duration:")
        self.duration_entry.pack(padx=(0, 20), side=tk.LEFT)

        self.end_entry = TimeEntry(time_frame, "End:")
        self.end_entry.pack(side=tk.LEFT)

        button_frame = ttk.Frame(dialog)
        button_frame.pack(side=tk.BOTTOM, pady=20)
        ok_btn = ttk.Button(button_frame, text="OK", command=self.on_ok)
        ok_btn.pack(side=tk.LEFT, padx=(0, 5))
        cancel_btn = ttk.Button(button_frame, text="Cancel", command=self.on_cancel)
        cancel_btn.pack(side=tk.LEFT)

        self.dialog = dialog
        self.selected_title = None
        self.selected_member = None
        self.selected_duration = None
        self.selected_start = None
        self.selected_end = None
        self.selected_fixed = None

        dialog.protocol("WM_DELETE_WINDOW", self.on_close)

    def set_row_contents(self, default_row):
        title, member, start, end, duration, fixed, instruction = default_row
        self._update_entry(self.title_entry, title)
        self._update_entry(self.member_entry, member)
        self._update_entry(self.instructions_entry, instruction)
        self.duration_entry.set_time(duration)
        self.start_entry.set_time(start)
        self.end_entry.set_time(end)
        self.fixed_combobox.set(fixed)

        self._update()

    def set_current_time_range(self, start, end):
        self.current_time_range_label.set_range(start, end)

    def set_prev_sec(self, prev_end_sec):
        self.prev_end_sec = prev_end_sec

    def on_fixed_selected(self, event=None):
        self._update()
        # when start is set, entry should be current start time
        current_start = self.current_time_range_label.start_label.cget("text")
        current_start_sec = time_format.time_str_to_seconds(current_start)
        self.start_entry.set_seconds(current_start_sec)

    def _update(self):
        fixed_code = self.fixed_combobox.get()
        if fixed_code == "duration":
            self.start_entry.set_seconds("")
            self.end_entry.set_seconds("")
            self.duration_entry.enabled()
            self.start_entry.disabled()
            self.end_entry.disabled()
        elif fixed_code == "start":
            self.start_entry.enabled()
            self.end_entry.enabled()
            self.duration_entry.enabled()

    def on_ok(self):
        if self._validate_time_entry() is False:
            return
        self.selected_title = self.title_entry.get()
        self.selected_member = self.member_entry.get()
        self.selected_instruction = self.instructions_entry.get()
        self.selected_duration = self.duration_entry.get_time()
        self.selected_start = self.start_entry.get_time()
        self.selected_end = self.end_entry.get_time()
        self.selected_fixed = self.fixed_combobox.get()
        self.dialog.destroy()

    def on_cancel(self):
        self.selected_title = None
        self.dialog.destroy()

    def on_close(self):
        self.selected_title = None
        self.dialog.destroy()

    def _update_entry(self, entry, value):
        entry.delete(0, tk.END)
        entry.insert(0, value)

    def _validate_time_entry(self):
        fixed = self.fixed_combobox.get()
        if fixed == "":
            return True
        time_entry_combination = [
            self.duration_entry.is_blank(),
            self.start_entry.is_blank(),
            self.end_entry.is_blank(),
        ]
        if all(time_entry_combination):
            self.msg_label.config(text="No time entry", foreground="red")
            return False

        duration = self.duration_entry.get_seconds()
        start = self.start_entry.get_seconds()
        end = self.end_entry.get_seconds()

        validate_result = True
        # basic validation
        if fixed == "start":
            if start == 0 and self.prev_end_sec != 0:
                error_msg = '"Start" should be set.'
                validate_result = False
            elif duration > 0 and end > 0:
                error_msg = '"Duration" or "End" (or both) should be blank.'
                validate_result = False
            elif start >= end and end > 0:
                error_msg = '"Start" should be smaller than "End".'
                validate_result = False
            else:
                error_msg = "OK"
                validate_result = True
        elif fixed == "duration":
            if duration == 0:
                error_msg = '"Duration" should be set.'
                validate_result = False
            else:
                error_msg = "OK"
                validate_result = True

        print(error_msg)
        if validate_result is False:
            self.msg_label.config(text=error_msg, foreground="red")
            return False
        else:
            if start == 0:
                start_sec = self.prev_end_sec
            else:
                start_sec = start
            if duration == 0 and end == 0:
                end_sec = start_sec
            elif duration > 0:
                end_sec = start_sec + duration
            else:
                end_sec = end

        if start_sec > end_sec:
            print("start > end")
            return False
        return True


class TimeEntry(ttk.Frame):
    def __init__(self, master, label_text: str):
        super().__init__(master)

        label = ttk.Label(self, text=label_text)
        label.pack(side=tk.LEFT)
        vcmd = (self.register(self._validate), "%P")
        invcmd = (self.register(self._invalid), "%P")
        self.time_entry = ttk.Entry(
            self, width=7, validatecommand=vcmd, invalidcommand=invcmd
        )
        self.time_entry.pack(side=tk.LEFT)

    def set_time(self, time_str):
        self.time_entry.delete(0, tk.END)
        self.time_entry.insert(0, time_str)

    def get_time(self):
        return self.time_entry.get()

    def set_seconds(self, seconds):
        if seconds == "":
            self.set_time("")
            return
        time_str = time_format.seconds_to_time_str(seconds)
        self.set_time(time_str)

    def get_seconds(self):
        time_str = self.time_entry.get()
        if time_str == "":
            return 0
        return time_format.time_str_to_seconds(time_str)

    def get_status(self):
        status = self.time_entry.cget("state")
        return status

    def enabled(self):
        self.time_entry.config(state="normal")

    def disabled(self):
        self.time_entry.config(state="disabled")

    def is_blank(self):
        return self.time_entry.get() == ""

    def _validate(self, text):
        p = r"(?:[0-5]?\d):[0-5]?\d"
        m = re.match(p, text)
        return m is not None

    def _invalid(self, text):
        self.time_entry.delete(0, tk.END)


class TimeRangeLabel(ttk.Frame):
    def __init__(self, master, label_str):
        super().__init__(master)
        title_label = ttk.Label(self, text=label_str)
        title_label.pack(side=tk.LEFT)

        self.start_label = ttk.Label(self, text="")
        self.start_label.pack(side=tk.LEFT)

        interval_label = ttk.Label(self, text="-")
        interval_label.pack(padx=5, side=tk.LEFT)

        self.end_label = ttk.Label(self, text="")
        self.end_label.pack(side=tk.LEFT)

    def set_range(self, start, end):
        self.start_label.config(text=start)
        self.end_label.config(text=end)



import threading
import tkinter as tk
from tkinter import ttk


class BleButtonManager:
    """BLEボタンとステータス表示を管理するクラス"""

    def __init__(
        self, parent_frame, master_window, trigger_device, ble_names, stop_delay_sec=1
    ):
        """
        Args:
            parent_frame: ボタンとラベルを配置する親フレーム
            master_window: UIの更新をスケジュールするためのメインウィンドウ
            trigger_device: BLE制御用のTriggerインスタンス
        """
        self.master_window = master_window
        self.trigger_device = trigger_device
        self.stop_delay_sec = stop_delay_sec
        self.trigger_device.set_delay_sec(self.stop_delay_sec)

        # BLE UI要素を作成
        self.ble_frame = ttk.Frame(parent_frame)
        self.ble_frame.pack(side=tk.RIGHT)

        self.ble_btn = ttk.Button(
            self.ble_frame, text="BLE Connect", command=self.connect_ble
        )
        self.ble_btn.pack(padx=0, side=tk.LEFT)

        self.ble_status_label = ttk.Label(self.ble_frame)
        self.ble_status_label.pack(padx=12, side=tk.LEFT)

        self.default_fg_color = self.ble_status_label.cget("foreground")

        if ble_names and len(ble_names) > 0:
            self.trigger_device.set_device_names(ble_names)
            dev_names = ", ".join(self.trigger_device.target_device_names)
            self.trigger_device.set_status(dev_names)
        else:
            self.trigger_device.set_status("No devices configured")
            self.ble_btn.config(state="disabled")
            self.ble_status_label.config(foreground="gray")

    def connect_ble(self):
        """BLE接続を開始"""
        self.set_disabled()
        self.trigger_device.set_status("Connecting...")

        def connect_thread():
            try:
                self.trigger_device.ble_connect()
                self.master_window.after(0, self._on_ble_connect_complete)
            except Exception as e:
                print(f"BLE connection error: {e}")
                self.master_window.after(0, self._on_ble_connect_error)

        thread = threading.Thread(target=connect_thread, daemon=True)
        thread.start()

    def set_disabled(self):
        self.ble_btn.config(state="disabled")

    def set_enabled(self):
        self.ble_btn.config(state="normal")

    def _on_ble_connect_complete(self):
        """BLE接続完了時にメインスレッドで呼ばれる"""
        self.ble_btn.config(state="normal")
        self.ble_status_label.config(text=self.trigger_device.get_status())

    def _on_ble_connect_error(self):
        """BLE接続エラー時にメインスレッドで呼ばれる"""
        self.ble_btn.config(state="normal")
        self.trigger_device.set_status("Connection Error")

    def update_ble_status(self):
        """BLEステータスラベルを更新"""
        status = self.trigger_device.update_status()
        self.ble_status_label.config(text=status)



import os

from . import csv_to_timetable


class FileLoader:
    def __init__(
        self,
        intermission_desc="Intermission",
    ):
        self.stage_list = []
        self.intermission_desc = intermission_desc
        self.encoding = None

    def _read_file(self, tar_path):
        self.stage_list = []
        if not os.path.exists(tar_path):
            print(f"File not found: {tar_path}")
            return None
        try:
            with open(tar_path, "r", encoding="utf-8") as f:
                self.encoding = "utf-8"
                return f.read()
        except UnicodeDecodeError:
            print("UTF-8 decoding failed, falling back to Shift-JIS")
            with open(tar_path, "r", encoding="shift-jis") as f:
                self.encoding = "shift-jis"
                return f.read()

    def load_file_for_preview(self, csv_path):
        timetable_csv_str = self._read_file(csv_path)
        if timetable_csv_str is None:
            return "Load file failed", None
        time_table = csv_to_timetable.TimeTable()
        try:
            warn_msg = time_table.load_csv_str(timetable_csv_str)
        except ValueError as e:
            raise e
        return warn_msg, time_table

    def load_file_for_timer(self, start_index: int, csv_path: str):
        timetable_csv_str = self._read_file(csv_path)
        if timetable_csv_str is None:
            return
        time_table = csv_to_timetable.TimeTable()
        time_table.load_csv_str(timetable_csv_str)

        start_row = time_table.get_timetable()[start_index]
        for i, row in enumerate(time_table.get_timetable()):
            if i < start_index:
                continue
            self.stage_list.append(
                {
                    "title": row["title"],
                    "start_dt": row["start"] - start_row["start"],
                    "end_dt": row["end"] - start_row["start"],
                    "duration": row["end"] - row["start"],
                    "member": row["member"],
                    "instruction": row["instruction"],
                }
            )

        # Intermission check
        intermission_list = []
        for i, stage in enumerate(self.stage_list):
            # check if intermission space exists
            current_end_dt = stage["end_dt"]
            if i + 1 < len(self.stage_list):
                next_start_dt = self.stage_list[i + 1]["start_dt"]
                if current_end_dt != next_start_dt:
                    intermission_list.append(
                        {
                            "title": self.intermission_desc,
                            "start_dt": current_end_dt,
                            "end_dt": next_start_dt,
                            "duration": next_start_dt - current_end_dt,
                            "member": "",
                            "instruction": "",
                        }
                    )
                else:
                    intermission_list.append(None)
        # Insert intermission into stage list from the end
        for i in range(len(intermission_list) - 1, -1, -1):
            if intermission_list[i] is not None:
                self.stage_list.insert(i + 1, intermission_list[i])

    def get_stage_list(self):
        return self.stage_list

    def get_encoding(self):
        return self.encoding

    def clear(self):
        self.stage_list = []


def utf8_to_sjis(tar_path: str) -> bool:
    try:
        with open(tar_path, "r", encoding="utf-8") as f:
            content = f.read()
    except UnicodeDecodeError as e:
        print(f"Error converting {tar_path}: {e}")
        return False

    if _can_encode_cp932(content):
        with open(tar_path, "w", encoding="cp932", errors="replace") as f:
            f.write(content)
    else:
        print(f"Content in {tar_path} cannot be encoded in cp932.")
        return False
    return True


def _can_encode_cp932(text: str) -> bool:
    try:
        text.encode("cp932")
        return True
    except UnicodeEncodeError:
        return False



import datetime

from . import time_format


class TimeTable:
    BOM = "\ufeff"

    def __init__(self):
        self.time_table = []
        self.current_time = 0

    def load_csv_str(self, csv_str):
        warn_msg = ""
        # UTF-8 BOM removal
        if csv_str.startswith(self.BOM):
            bom_len = len(self.BOM)
            csv_str = csv_str[bom_len:]

        lines = csv_str.strip().split("\n")

        # Validate the header
        header = lines[0].strip().split(",")
        # header-correct_header dictionary to get the index of each header
        header_dict = {header[i]: i for i in range(len(header))}

        # Process each row
        if "end" not in header_dict.keys():
            is_no_end = True
            header_dict["end"] = -1
        else:
            is_no_end = False
        # For checking if the previous row has an end time by "end_sec_str" or "duration_sec_str"
        has_end_time = True
        for i, line in enumerate(lines[1:]):
            # Skip if the line is empty like ",,,,," or ",,,,"
            if line.strip().replace(",", "") == "":
                print(f"[line {i + 1}] Empty line")
                continue

            (
                title,
                member,
                duration_sec_str,
                start_sec_str,
                end_sec_str,
                fixed,
                instruction,
            ) = self._asign(line, header_dict, is_no_end)

            duration_sec = time_format.time_str_to_seconds(duration_sec_str)
            start_sec = time_format.time_str_to_seconds(start_sec_str)
            end_sec = time_format.time_str_to_seconds(end_sec_str)
            if fixed not in ["start", "duration", "none"]:
                raise ValueError(f"[line {i + 1}] Invalid fixed code: {fixed}")

            has_error = False
            if fixed == "start":
                if start_sec < self.current_time:
                    warn_msg = f"[line {i + 1}] {title} Conflict with the previous line"
                    has_error = True

                if start_sec == 0 and i != 0:
                    raise ValueError(
                        f"[line {i + 1}] start_sec must be set in fixed==start"
                    )
                if duration_sec > 0:
                    end_sec = start_sec + duration_sec
                elif end_sec > 0:
                    end_sec = end_sec
                # if duration and end are not set, get the next start time
                elif i < len(lines) - 2:
                    next_line = lines[i + 2]
                    end_sec = self.get_next_start(line, next_line, header_dict)
                    if end_sec == 0:
                        end_sec = start_sec
                else:
                    raise ValueError(f"[line {i + 1}] No next line")
            elif fixed == "duration":
                if duration_sec == 0:
                    raise ValueError(
                        f"[line {i + 1}] duration_sec must be set in fixed==duration"
                    )
                if start_sec == 0 and end_sec == 0:
                    start_sec = self.current_time
                    end_sec = start_sec + duration_sec
                elif start_sec > 0:
                    start_sec = start_sec
                    end_sec = start_sec + duration_sec
                elif end_sec > 0:
                    end_sec = end_sec
                    start_sec = end_sec - duration_sec
                if (i > 0) and (has_end_time is False):
                    warn_msg = (
                        f"[line {i + 1}] No Duration (or End) in the previous line"
                    )
                    has_error = True

            start_td = datetime.timedelta(seconds=start_sec)
            end_td = datetime.timedelta(seconds=end_sec)

            # Add task start and end times to the timetable
            self.time_table.append(
                {
                    "title": title,
                    "start": start_td,
                    "end": end_td,
                    "member": member,
                    "duration_sec": duration_sec_str,
                    "start_sec": start_sec_str,
                    "end_sec": end_sec_str,
                    "fixed": fixed,
                    "instruction": instruction,
                    "has_error": has_error,
                }
            )

            # Update current time
            self.current_time = end_sec
            has_end_time = end_sec_str != "" or duration_sec_str != ""
        return warn_msg

    def _asign(self, line_str, header_dict, is_no_end=False):
        splited_line = line_str.strip().split(",")
        title = splited_line[header_dict["title"]]
        member = splited_line[header_dict["member"]]
        duration_sec_str = splited_line[header_dict["duration"]]
        start_sec_str = splited_line[header_dict["start"]]
        if is_no_end:
            end_sec_str = ""
        else:
            end_sec_str = splited_line[header_dict["end"]]
        fixed = splited_line[header_dict["fixed"]]
        if "instruction" in header_dict.keys():
            instruction = splited_line[header_dict["instruction"]]
        else:
            instruction = ""
        return (
            title,
            member,
            duration_sec_str,
            start_sec_str,
            end_sec_str,
            fixed,
            instruction,
        )

    def get_timetable(self):
        return self.time_table

    def get_timetable_as_str(self):
        ret_table = []
        for entry in self.time_table:
            start_str = time_format.timedelta_to_str(entry["start"])
            end_str = time_format.timedelta_to_str(entry["end"])
            ret_table.append(
                {
                    "title": entry["title"],
                    "start": start_str,
                    "end": end_str,
                    "member": entry["member"],
                }
            )
        return ret_table

    def get_next_start(self, current_line, next_line, header_dict):
        (
            title,
            member,
            duration_sec_str,
            start_sec_str,
            end_sec_str,
            fixed,
            instruction,
        ) = self._asign(next_line, header_dict)
        if start_sec_str == "0":
            raise ValueError(f"next_start_sec is 0: {current_line}")
        return time_format.time_str_to_seconds(start_sec_str)



import os
import shutil


def check_file_locked(file_path):
    try:
        with open(file_path, "r+") as f:
            pass
    except PermissionError:
        return True
    return False


def move_to_backup_folder(file_path):
    backup_dir = "backup"
    os.makedirs(backup_dir, exist_ok=True)
    backup_file = os.path.join(backup_dir, os.path.basename(file_path))
    if os.path.exists(file_path):
        shutil.move(file_path, backup_file)



def save_as_svg(tk_canvas, filename):
    svg = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n'
    width = tk_canvas.winfo_width() + 100
    height = tk_canvas.winfo_height()
    svg += f"<svg xmlns='http://www.w3.org/2000/svg' width='{width}' height='{height}' viewBox='0 0 {width} {height}'>\n"

    for item in tk_canvas.find_all():
        if tk_canvas.type(item) == "rectangle":
            x1, y1, x2, y2 = tk_canvas.coords(item)
            color = tk_canvas.itemcget(item, "fill")
            svg += f"<rect x='{x1}' y='{y1}' width='{x2-x1}' height='{y2-y1}' fill='{color}' stroke='black' />\n"
        elif tk_canvas.type(item) == "text":
            x, y = tk_canvas.coords(item)
            text = tk_canvas.itemcget(item, "text")
            anchor = tk_canvas.itemcget(item, "anchor")
            font = tk_canvas.itemcget(item, "font")
            font_size = int(font.split(" ")[1])

            if anchor == "e":
                anchor = "end"
            elif anchor == "w":
                anchor = "start"
            elif anchor == "n":
                anchor = "middle"

            y += font_size / 2 - 2
            svg += f"<text x='{x}' y='{y}' font-family='Helvetica' font-size='{font_size}' text-anchor='{anchor}'>{text}</text>\n"
        elif tk_canvas.type(item) == "line":
            x1, y1, x2, y2 = tk_canvas.coords(item)
            svg += f"<line x1='{x1}' y1='{y1}' x2='{x2}' y2='{y2}' stroke='black' />\n"

    svg += "</svg>"
    with open(filename, "w", encoding="utf-8") as f:
        f.write(svg)
